print("import: u'cherrypy'")
import cherrypy

print("import: u'cherrypy.lib'")
import cherrypy.lib

print("import: u'cherrypy.process'")
import cherrypy.process

print("import: u'cherrypy.scaffold'")
import cherrypy.scaffold

print("import: u'cherrypy.test'")
import cherrypy.test

print("import: u'cherrypy.tutorial'")
import cherrypy.tutorial

